import datetime
import time
import os

from zkcluster import util


class CSVDumper(object):
    def __init__(self, server):
        self.output_file = server.service_config['csv_dump']
        self.async_suite = server.async_suite
        self.stat = server.memos['stats']
        self.columns = server.service_config['csv_columns']
        self.token = server.service_config['csv_token']

        file_exists = os.access(self.output_file, os.F_OK)
        self.stream = open(self.output_file, 'w+')
        if not file_exists:
            self.stream.write(
                ",".join(self.columns) + "\n"
            )
            self.stream.flush()

        self.worker = self.async_suite.background_thread(self.dump)
        util.event_listen(self.stat, "message_received", self.message_received)
        self.last_updated = time.time()

    def message_received(self, message):
        self.last_updated = time.time()

    def dump(self):
        # TOOD: allow options for event-only dumping, formats besides
        # CSV(?), file rotation(?), per-server and per-process dumps
        # (e.g. not just per-application)
        last_written = time.time()
        while True:
            self.async_suite.sleep(1)
            if self.last_updated <= last_written:
                continue

            last_written = time.time()
            timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

            hostprogs = list(self.stat.hostprogs.values())
            hostprogs.sort(
                key=lambda hostprog: (hostprog.hostname, hostprog.progname)
            )
            for hostprog in hostprogs:
                row = {
                    'timestamp': timestamp,
                    'token': self.token, 'hostname': hostprog.hostname,
                    'progname': hostprog.progname,
                    'process_count': hostprog.process_count,
                    'connection_count': hostprog.connection_count,
                    'max_process_count': hostprog.max_process_count,
                    'max_checkedout': hostprog.max_checkedout,
                    'max_connections': hostprog.max_connections,
                    'checkout_count': hostprog.checkout_count
                }

                self.stream.write(
                    ",".join(str(row.get(col)) for col in self.columns) + "\n"
                )
            self.stream.flush()



import os
import socket
import sys
import thread
import time
import re
import logging

from sqlalchemy import event
from sqlalchemy.engine import url as _url

from . import client
from . import stats
from zkcluster import util

log = logging.getLogger(__name__)


class CMWrapper(object):
    @classmethod
    def get_dialect_cls(cls, url):
        new_url = _url.URL(re.sub(r'.connmon', '', url.drivername))

        addr = url.query.pop('connmon_addr', None)
        service = url.query.pop('connmon_service', None)
        config = url.query.pop('connmon_config', None)
        if addr:
            host, port = addr.split(':')
            url._connmon_info = {"host": host, "port": port, "config": config}
        elif service:
            url._connmon_info = {"service": service, "config": config}
        else:
            url._connmon_info = {"service": "default", "config": config}
        return new_url.get_dialect()

    @classmethod
    def engine_created(cls, engine):
        connect_info = engine.url._connmon_info
        ConnectionPoolAnalyzer(connect_info, engine)


class ConnectionPoolAnalyzer(object):

    def __init__(self, connect_info, engine):
        self.connect_info = connect_info
        self.async_suite = util.async_suite(green=False)
        self.lock = self.async_suite.lock()
        self.hostname = socket.gethostname()
        self.progname = sys.argv[0]
        self.ppid = os.getppid()
        self.pid = os.getpid()
        self.client = None

        event.listen(engine, "connect", self._connect_evt)
        event.listen(engine, "checkout", self._checkout_evt)
        event.listen(engine, "checkin", self._checkin_evt)
        event.listen(engine, "invalidate", self._invalidate_evt)
        event.listen(engine, "soft_invalidate", self._invalidate_evt)

        self._patch_pool(engine.pool)

    def _get_client(self):
        pid = os.getpid()

        if self.client is not None and pid == self.pid:
            return self.client

        with self.lock:
            if self.client is None or pid != self.pid:
                # adjust for new subprocess
                self.pid = pid
                self.client = client.StatsClient.from_connect_info(
                    self.connect_info,
                    self.hostname, self.progname, self.pid
                )
            return self.client

    def _connect_evt(self, dbapi_conn, connection_rec):
        self._send_evt(dbapi_conn, stats.CONNECT)

    def _checkout_evt(self, dbapi_conn, connection_rec, connection_proxy):
        self._send_evt(dbapi_conn, stats.CHECKOUT)

    def _checkin_evt(self, dbapi_conn, connection_rec):
        self._send_evt(dbapi_conn, stats.CHECKIN)

    def _invalidate_evt(self, dbapi_conn, connection_rec, exception):
        self._send_evt(dbapi_conn, stats.INVALIDATE)

    def _close_connection(self, dbapi_conn):
        self._pool_close_connection(dbapi_conn)
        self._send_evt(dbapi_conn, stats.CLOSE)

    def _send_evt(self, dbapi_conn, action):
        client = self._get_client()

        evt = stats.ConnectionEvent(
            self.hostname, self.progname, self.pid, self.ppid,
            thread.get_ident(), repr(dbapi_conn), time.time(), action
        )
        client.send(evt)

    def _recreate_pool(self):
        recreated = self._pool_recreate()
        self._patch_pool(recreated)
        return recreated

    def _patch_pool(self, pool):
        self._pool_close_connection = pool._close_connection
        self._pool_recreate = pool.recreate
        pool._close_connection = self._close_connection
        pool.recreate = self._recreate_pool

    def close(self):
        if self.client is not None:
            self.client.close()
            self.client = None




import curses
import functools
import operator
import re
import time
import sys

from zkcluster import util


COLOR_MAP = {
    "K": curses.COLOR_BLACK,
    "R": curses.COLOR_RED,
    "B": curses.COLOR_BLUE,
    "C": curses.COLOR_CYAN,
    "G": curses.COLOR_GREEN,
    "M": curses.COLOR_MAGENTA,
    "R": curses.COLOR_RED,
    "W": curses.COLOR_WHITE,
    "Y": curses.COLOR_YELLOW,
    "D": -1
}

_TEXT_RE = re.compile(r'(#.+?)&', re.M)


class Display(object):
    def __init__(self, stat, client):
        self.columns = [
            ('hostname (#R&[dis]#G&connected#d&)', "%s", .20, "L"),
            ('progname', "%s", .20, "L"),
            ('nproc', "%d", .08, "R"),
            ('conn', "%d", .08, "R"),
            ('ckout', "%d", .08, "R"),
            ('maxnproc', "%d", .12, "R"),
            ('maxconn', "%d", .12, "R"),
            ('maxckout', "%d", .12, "R"),
        ]
        self.stat = stat
        self.client = client  # zkconfig.client.LocalClient

        self._winsize = None
        self._x_positions = None

    def _calc_x_positions(self):
        x = 0
        widths = []
        midline = False
        for idx, col in enumerate(self.columns):
            cname = col[0]
            width = col[2]
            just = col[3]
            charwidth = max(len(cname), int(self._winsize[1] * width))
            if just == "L":
                widths.append((x, charwidth))
                x += charwidth
            else:
                width = sum(
                    max(len(col[0]), int(self._winsize[1] * col[2]))
                    for col in self.columns[idx:]
                )
                x = self._winsize[1] - width
                if not midline:
                    midline = True
                    midline_at = widths[-1][0] + widths[-1][1]
                    midline_over = midline_at - x
                    if midline_over > 0:
                        widths[-1] = \
                            widths[-1][0], widths[-1][1] - midline_over
                widths.append((x, charwidth))

        self._x_positions = widths

    def _refresh_winsize(self):
        old_winsize = self._winsize

        self._winsize = self.window.getmaxyx()
        if old_winsize != self._winsize or \
                curses.is_term_resized(*old_winsize):
            curses.resize_term(*self._winsize)
            self._calc_x_positions()

    def start(self):
        self.enabled = True
        window = curses.initscr()

        curses.noecho()
        window.erase()
        window.nodelay(1)
        curses.start_color()
        curses.use_default_colors()
        self._color_pairs = {}
        for i, (k, v) in enumerate(COLOR_MAP.items(), 1):
            curses.init_pair(i, v, -1)
            self._color_pairs[k] = curses.color_pair(i)
        self._color_pairs['b'] = curses.A_BOLD
        self._color_pairs['n'] = curses.A_NORMAL
        window.refresh()
        self.window = window
        self._refresh_winsize()

        try:
            with util.stop_on_keyinterrupt():
                self._redraw()
        finally:
            self.stop()

    def _redraw(self):
        render_timer = util.periodic_timer(.5)
        while self.enabled:
            time.sleep(.1)
            if render_timer(time.time()):
                self._render()
            self._handle_cmds()

    def _handle_cmds(self):
        char = self.window.getch()
        if char in (ord('Q'), ord('q')):
            self.stop()
        elif char == curses.KEY_RESIZE:
            # NOTE: this char breaks if you import readline, which
            # is implicit if you use Python cmd.Cmd() in its default
            # mode
            self._refresh_winsize()

    def stop(self):
        self.enabled = False
        curses.endwin()

    def _get_color(self, color):
        try:
            return self._color_pairs[color]
        except KeyError:
            assert len(color) > 1
            mapped = functools.reduce(
                operator.or_,
                [self._color_pairs[char] for char in color]
            )
            self._color_pairs[color] = mapped
            return mapped

    def _render_str(self, y, x, text, default_color="D", max_=None):
        if x < 0:
            x = self._winsize[1] - len(_TEXT_RE.sub('', text))

        current_color = dflt = self._get_color(default_color)
        if max_:
            max_x = x + max_
        else:
            max_x = self._winsize[1]
        for token in _TEXT_RE.split(text):
            if token.startswith("#"):
                ccode = token[1:]
                if ccode == "d":
                    current_color = dflt
                else:
                    current_color = self._get_color(ccode)
            else:
                self.window.addstr(y, x, token[:max_x - x], current_color)
                x += len(token)
                if x > max_x:
                    break

    def _render_row(self, row, y):
        x_positions = iter(self._x_positions)
        for elem, col in zip(row, self.columns):
            cname, fmt, width, justify = col
            elem = fmt % (elem, )
            x, charwidth = next(x_positions)
            self._render_str(y, x, elem, max_=charwidth - 1)

    def _render(self):
        self.window.erase()

        self._render_str(
            0, 0,
            "#Bb&[Connmon]#Dn& [%s:%s | %s:%s]" % (
                self.client.servicename,
                self.client.nodename, self.client.host, self.client.port
            )
        )
        self._render_str(
            0, -1,
            "#D&Commands: #Y&(Q)#D&uit"
        )

        self._render_str(
            2, 0,
            "#Mb&Hosts: #Dn&[%d / %d max]  #Mb&Processes: #Dn&[%d / %d max]  "
            "#Mb&Connections: #Dn&[%d / %d max]  "
            "#Mb&Checkouts: #Dn&[%d / %d max]" % (
                self.stat.host_count,
                self.stat.max_host_count,
                self.stat.process_count,
                self.stat.max_process_count,
                self.stat.connection_count,
                self.stat.max_connection_count,
                self.stat.checkout_count,
                self.stat.max_checkout_count
            ),
            "Wb"
        )

        top = 5

        x_positions = iter(self._x_positions)
        for col in self.columns:
            cname, fmt, width, justify = col
            x, charwidth = next(x_positions)
            self._render_str(top, x, cname, "Cb")

        rows = []
        hostprogs = list(self.stat.hostprogs.values())
        hostprogs.sort(
            key=lambda hostprog: (hostprog.hostname, hostprog.progname)
        )
        for hostprog in hostprogs:
            is_connected = bool(hostprog.process_count)
            rows.append((
                "#%s&%s" % (
                    "G" if is_connected else "R",
                    hostprog.hostname,
                ),
                "#%s&%s" % (
                    "G" if is_connected else "R",
                    hostprog.short_progname,
                ),
                hostprog.process_count,
                hostprog.connection_count,
                hostprog.checkout_count,
                hostprog.max_process_count,
                hostprog.max_connections, hostprog.max_checkedout
            ))

        for y, row in enumerate(rows, top + 1):
            self._render_row(row, y)
        self.window.refresh()




from zkcluster import rpc
from zkcluster import p2p
from zkcluster import util

import logging
from . import stats
from . import dumper

log = logging.getLogger(__name__)


def init_server(server):
    stat_ = stats.QueuedStats(_green=True)
    server.speak_rpc(rpc_reg)
    server.memos['stats'] = stat_
    server.memos['consoles'] = consoles = set()
    p2p_ = p2p.PeerRunner(server)
    p2p_.speak_rpc(rpc_reg)

    def message_received(message):
        # broadcast stats events to listening consoles

        msg = ConsoleStatsMessage(message)
        for client in consoles:
            msg.send(client.rpc_service)

    def new_peer(peer):
        peer.memos['stats'] = stat_

    def peer_connected(peer):
        evt = stats.SummaryEvent(stat_.summary(), stats.SUMMARY)
        msg = P2PStatsMessage(evt)
        msg.send(peer.rpc_service)

    def start_stats(server):
        stat_.start()

    util.event_listen(server, "before_listen", start_stats)
    util.event_listen(
        server.memos['stats'], "message_received", message_received)
    util.event_listen(p2p_, "new_peer", new_peer)
    util.event_listen(p2p_, "peer_connected", peer_connected)
    util.event_listen(
        server, "client_connection_disconnected", client_disconnected)

    if server.service_config.get('csv_dump'):
        dumper.CSVDumper(server)

    return server


def client_disconnected(server, client_connection, unexpected, message):
    if 'client_info' in client_connection.memos:
        log.info("Stats client %s disconnected", client_connection)
        info = client_connection.memos['client_info']

        stats = info['stats']
        evt = stats.process_closed(
            info['remote_hostname'],
            info['remote_progname'],
            info['remote_pid']
        )

        msg = P2PStatsMessage(evt)
        p2p_ = client_connection.server.memos['p2p']
        p2p_.broadcast_rpc_to_peers(msg)

    elif 'console' in client_connection.memos:
        log.info("Console client %s disconnected", client_connection)
        server.memos['consoles'].remove(client_connection)


rpc_reg = rpc.RPCReg()


@rpc_reg.call("remote_hostname", "remote_progname", "remote_pid", "summary")
class InitClientRPC(rpc.RPCEvent):
    """Initial rpc call sent from a database client to the server.

    Identifies the client as a database listener which will send us events.

    Request is processed server-side.

    """

    def receive_request(self, rpc, service_connection):
        log.info("New stats client %s", service_connection)

        stats = service_connection.server.memos['stats']
        service_connection.memos['client_info'] = {
            'remote_hostname': self.remote_hostname,
            'remote_progname': self.remote_progname,
            'remote_pid': self.remote_pid,
            'stats': stats
        }

        evt = stats.summary_added(self.summary)

        msg = P2PStatsMessage(evt)
        p2p_ = service_connection.server.memos['p2p']
        p2p_.broadcast_rpc_to_peers(msg)


@rpc_reg.call("payload")
class StatsMessage(rpc.RPCEvent):
    """Send a stats event from a database client to the server.

    Request is processed server-side.

    """
    def receive_request(self, rpc, service_connection):
        stats_ = service_connection.memos['client_info']['stats']
        stats_.receive_message(
            stats.evt_from_tuple(self.payload)
        )

        msg = P2PStatsMessage(self.payload)
        p2p_ = service_connection.server.memos['p2p']
        p2p_.broadcast_rpc_to_peers(msg)


@rpc_reg.call("payload")
class P2PStatsMessage(rpc.RPCEvent):
    """Send a stats event from a server peer to another server peer

    Request is processed server-side.

    """
    def receive_request(self, rpc, service_connection):
        stats_ = service_connection.memos['p2p_peer'].memos['stats']
        stats_.receive_message(
            stats.evt_from_tuple(self.payload)
        )


@rpc_reg.call()
class InitConsoleRPC(rpc.RPC):
    """Initial rpc call sent from a console client to the server.

    Identifies the client as a console, which will receive events from
    the server.

    Request is processed server-side.

    """

    def receive_request(self, rpc, service_connection):
        log.info("New console client %s", service_connection)

        server = service_connection.server   # RemoteServer instance

        service_connection.memos['console'] = True
        consoles = server.memos['consoles']
        consoles.add(service_connection)

        stats = server.memos['stats']
        return stats.summary()


@rpc_reg.call("payload")
class ConsoleStatsMessage(rpc.RPCEvent):
    """Send a stats event from the server to a listening console.

    Request is processed client-side.

    """
    def receive_request(self, rpc, console):
        with console.memos['stats'].receive_mutex:
            console.memos['stats'].receive_message(
                stats.evt_from_tuple(self.payload)
            )




import os
from . import stats
import logging
from . import server
from . import frontend as _frontend
from zkcluster import client as _client
from zkcluster import util
log = logging.getLogger(__name__)


class StatsClient(object):
    def __init__(self, client, hostname, progname, pid):
        self.hostname = hostname
        self.progname = progname
        self.pid = pid
        self.connected = False
        self.local_state = stats.Stats()
        self.local_state.process_added(hostname, progname, pid)
        self.async_suite = util.async_suite(green=False)
        self.send_mutex = self.async_suite.lock()

        self.client = client
        self.rpc_service = self.client.speak_rpc(server.rpc_reg)
        util.event_listen(self.client, "client_connected", self.on_connect)
        util.event_listen(
            self.client, "client_disconnected", self.on_disconnect)
        self.client.connect_persistent()

    @classmethod
    def from_connect_info(cls, connect_info, hostname, progname, pid):

        if 'service' in connect_info:
            servicename = connect_info['service']
            config = _frontend.load_config(connect_info['config'])
            service_config = config.config_for_servicename(
                connect_info['service'])
            client = _client.LocalClient.from_config(
                service_config, servicename, receive_events=False)
        elif 'host' in connect_info:
            host = connect_info['host']
            port = connect_info['port']
            client = _client.LocalClient.from_host_port(
                None, None, host, port, receive_events=False)
        else:
            raise ValueError("'service' or 'host' expected")

        return StatsClient(client, hostname, progname, pid)

    def on_connect(self, client):
        server.InitClientRPC(
            self.hostname,
            self.progname,
            self.pid,
            self.local_state.summary()
        ).send(self.rpc_service)
        self.connected = True

    def on_disconnect(self, client):
        self.connected = False

    def send(self, evt):
        assert self.pid == os.getpid()

        with self.send_mutex:
            self.local_state.receive_message(evt)
            # note this is different from self.client.connected
            if self.connected:
                server.StatsMessage(evt).send(self.rpc_service)

    def close(self):
        raise NotImplementedError()



import os
import logging

from zkcluster import config as _config
from zkcluster import client as _client
from zkcluster import cmdline as _cmdline
from zkcluster import exc
from zkcluster import util

from . import stats as _stats
from . import server as _server
from . import display as _display


_config_elements = [
    _config.ConfigElement("csv_dump"),
    _config.ConfigElement("csv_token", default="connmon"),
    _config.ConfigElement("csv_columns", _config.StringListFixer, default=[
        "timestamp", "token", "hostname", "progname", "process_count",
        "connection_count", "checkout_count", "max_process_count",
        "max_connections", "max_checkedout"
    ])
]


def load_config(config_file=None, logging_=True):
    if config_file:
        cfg_file = config_file
    else:
        cfg_file = "/etc/connmon.cfg"
        if not os.access(cfg_file, os.F_OK):
            cfg_file = None

    if cfg_file:
        config = _config.Config.from_config_file(
            cfg_file, prefix="connmon_", config_elements=_config_elements)
        if logging_:
            config.load_logging_configs()
    else:
        config = _config.Config.from_config_string("""
[connmon_service_default]
name: default

nodes:
    node1 hostname=localhost:5800 bind=0.0.0.0
""", prefix="connmon_", config_elements=_config_elements)
        if logging_:
            logging.basicConfig()
            logging.getLogger("connmon").setLevel(logging.INFO)
            logging.getLogger("zkcluster").setLevel(logging.INFO)

    return config


class SingleCmdLine(_cmdline.SingleCmdLine):

    def load_config(self):
        if self.options.logging:
            self.cmd.logging = True

        self.config = load_config(
            self.options.config, logging_=self.cmd.logging)


class ListenCmd(_cmdline.ListenCmd):
    logging = True

    def init_server(self, server):
        _server.init_server(server)


def init_client(client):
    client.speak_rpc(_server.rpc_reg)


def client_connected(client):
    stats = _stats.QueuedStats(_green=False)
    with stats.receive_mutex:
        client.memos['stats'] = stats
        summary = _server.InitConsoleRPC().send(client.rpc_service)
        stats.summary_added(summary)
    stats.start()


class MonitorCmd(_cmdline.ClientCmd):
    logging = False

    def create_subparser(self, parser, subparsers):
        return subparsers.add_parser(
            "monitor",
            help="run live monitor")

    def setup_arguments(self, parser):
        super(MonitorCmd, self).setup_arguments(parser)
        parser.add_argument(
            "--dump-only", action="store_true",
            help="dump messages only")
        parser.add_argument(
            "--logging", action="store_true",
            help="enable logging")
        return parser

    def go(self, cmdline):

        args = self.get_client_args(cmdline.options, cmdline.config)

        if "hostname" in args:

            client = _client.LocalClient.from_host_port(
                args['user'], args['password'],
                args['hostname'], args['port'])

        elif "service" in args:
            client = _client.LocalClient.from_config(
                args['service_config'], args['service'],
                args['node'], args['user'], args['password'])

        init_client(client)

        try:
            client.connect()
        except exc.AuthFailedError:
            print("auth failed")
        except exc.DisconnectedError as de:
            print(de)
        else:
            client_connected(client)

            if cmdline.options.dump_only:
                self._run_dumper(client)
            else:
                self._run_monitor(client)

    def _run_monitor(self, client):
        def on_disconnect(client):
            self.display.stop()
            print("Client connection lost")

        util.event_listen(
            client, "client_disconnected", on_disconnect)
        stats = client.memos['stats']
        self.display = _display.Display(stats, client)
        self.display.start()
        client.close()

    def _run_dumper(self, client):
        def message_received(msg):
            print("Message: %s" % (msg, ))

        def on_disconnect(self, client):
            self.display.stop()
            print("Client connection lost")

        stats = client.memos['stats']
        util.event_listen(stats, "message_received", message_received)

        import time
        self.connected = True
        while self.connected:
            time.sleep(1)


def monitor(argv=None):
    SingleCmdLine(MonitorCmd()).main(argv)


def connmond(argv=None):
    SingleCmdLine(ListenCmd()).main(argv)



import collections
from zkcluster import util
import os
import logging
import Queue


log = logging.getLogger(__name__)

ROLLBACK_ON_RETURN = 0
COMMIT_ON_RETURN = 1
CONNECT = 2
CLOSE = 3
CHECKOUT = 4
CHECKIN = 5
INVALIDATE = 6
RECYCLE = 7

PROCESS_ADDED = 8
PROCESS_REMOVED = 9
SUMMARY = 10

ConnectionEvent = collections.namedtuple(
    'ConnectionEvent', [
        'hostname', 'progname', 'pid', 'ppid',
        'tid', 'connection_id', 'timestamp', 'action'
    ]
)

ProcessEvent = collections.namedtuple(
    'ProcessEvent', [
        'remote_hostname', 'remote_progname', 'remote_pid',
        'action'
    ]
)

SummaryEvent = collections.namedtuple(
    'SummaryEvent', ['summary', 'action']
)


_lookup = ([ConnectionEvent] * 8) + ([ProcessEvent] * 2) + [SummaryEvent]


def evt_from_tuple(msg):
    return _lookup[msg[-1]](*msg)


class ConnectionStat(object):
    __slots__ = (
        'key', 'hostname', 'pid', 'connection_id',
        'checkedout', 'progname'
    )

    def __init__(self, hostname, pid, connection_id, checkedout, progname):
        self.key = "%s|%s|%s" % (hostname, pid, connection_id)
        self.hostname = hostname
        self.pid = pid
        self.connection_id = connection_id
        self.checkedout = checkedout
        self.progname = progname

    def summary(self):
        return (
            self.hostname, self.pid,
            self.connection_id, self.checkedout, self.progname)

    @classmethod
    def from_summary(self, summary):
        return ConnectionStat(*summary)

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return self.key == other.key


class HostProg(object):
    def __init__(self, hostname, progname):
        self.hostname = hostname
        self.progname = progname
        self.connections = {}
        self.checkedout = set()
        self.max_connections = 0
        self.max_checkedout = 0
        self.max_process_count = 0
        self.pids = set()

    def summary(self):
        return {
            "hostname": self.hostname,
            "progname": self.progname,
            "connections": dict(
                (key, conn.summary())
                for (key, conn) in self.connections.items()
            ),
            "checkedout": [conn.key for conn in self.checkedout],
            "max_connections": self.max_connections,
            "max_checkedout": self.max_checkedout,
            "max_process_count": self.max_process_count,
            "pids": list(self.pids)
        }

    def merge_summary(self, summary):
        self.connections.update(
            (key, ConnectionStat.from_summary(stat_summary))
            for key, stat_summary in summary['connections'].items(),
        )
        self.checkedout.update(
            self.connections[key] for key in summary['checkedout']
        )
        self.pids.update(summary['pids'])
        self.max_connections = max(
            summary['max_connections'],
            self.max_connections, len(self.connections))
        self.max_checkedout = max(
            summary['max_checkedout'],
            self.max_checkedout, len(self.checkedout))
        self.max_process_count = max(
            summary['max_process_count'],
            self.max_process_count, self.process_count)

    def connect(self, stat):
        self.connections[stat.key] = stat
        self.max_connections = max(self.max_connections, len(self.connections))

    def close(self, stat):
        del self.connections[stat.key]

    def checkout(self, stat):
        self.checkedout.add(stat)
        self.max_checkedout = max(self.max_checkedout, len(self.checkedout))

    def checkin(self, stat):
        self.checkedout.remove(stat)

    def invalidate(self, stat):
        self.checkedout.remove(stat)

    @property
    def short_progname(self):
        if self.progname:
            return os.path.basename(self.progname)
        else:
            return "<none>"

    @property
    def connection_count(self):
        return len(self.connections)

    @property
    def checkout_count(self):
        return len(self.checkedout)

    @property
    def process_count(self):
        return len(self.pids)

    def add_pid(self, remote_pid):
        self.pids.add(remote_pid)
        self.max_process_count = max(self.max_process_count, len(self.pids))

    def remove_pid(self, remote_pid):
        for key, value in list(self.connections.items()):
            if value.pid == remote_pid:
                del self.connections[key]
        self.pids.remove(remote_pid)


class Stats(object):
    def __init__(self, _green=False):
        self.hostprogs = {}
        self.async_suite = util.async_suite(green=_green)
        self.lock = self.async_suite.lock()

    @property
    def process_count(self):
        return sum(
            hostprog.process_count for hostprog in self.hostprogs.values()
        )

    @property
    def host_count(self):
        return len(
            set(hostprog.hostname
                for hostprog in self.hostprogs.values() if hostprog.pids)
        )

    @property
    def max_host_count(self):
        return len(
            set(hostprog.hostname for hostprog in self.hostprogs.values())
        )

    @property
    def max_process_count(self):
        return sum(
            hostprog.max_process_count for hostprog in self.hostprogs.values()
        )

    @property
    def connection_count(self):
        return sum(
            hostprog.connection_count for hostprog in self.hostprogs.values()
        )

    @property
    def checkout_count(self):
        return sum(
            hostprog.checkout_count for hostprog in self.hostprogs.values()
        )

    @property
    def max_connection_count(self):
        return sum(
            hostprog.max_connections for hostprog in self.hostprogs.values()
        )

    @property
    def max_checkout_count(self):
        return sum(
            hostprog.max_checkedout for hostprog in self.hostprogs.values()
        )

    def summary(self):
        return [host.summary() for host in self.hostprogs.values()]

    def _get_hostprog(self, remote_hostname, remote_progname):
        with self.lock:
            try:
                hostprog = self.hostprogs[(remote_hostname, remote_progname)]
            except KeyError:
                hostprog = \
                    self.hostprogs[(remote_hostname, remote_progname)] = \
                    HostProg(remote_hostname, remote_progname)
        return hostprog

    def summary_added(self, summary):
        evt = SummaryEvent(
            summary, SUMMARY
        )
        self.receive_message(evt)
        return evt

    def process_added(self, remote_hostname, remote_progname, remote_pid):
        evt = ProcessEvent(
            remote_hostname, remote_progname, remote_pid,
            PROCESS_ADDED
        )
        self.receive_message(evt)
        return evt

    def process_closed(self, remote_hostname, remote_progname, remote_pid):
        evt = ProcessEvent(
            remote_hostname, remote_progname, remote_pid,
            PROCESS_REMOVED
        )

        self.receive_message(evt)
        return evt

    def _process_added(self, message):
        hostprog = self._get_hostprog(
            message.remote_hostname, message.remote_progname)
        hostprog.add_pid(message.remote_pid)

    def _process_closed(self, message):
        try:
            hostprog = self.hostprogs[
                (message.remote_hostname, message.remote_progname)]
        except KeyError:
            pass
        else:
            hostprog.remove_pid(message.remote_pid)

    def _summary_added(self, message):
        summary = message.summary
        for hostprog_summary in summary:
            remote_hostname, remote_progname = \
                hostprog_summary['hostname'], hostprog_summary['progname']

            hostprog = self._get_hostprog(remote_hostname, remote_progname)
            hostprog.merge_summary(hostprog_summary)

    def receive_message(self, message):
        log.debug("message received: %s", message)

        self.dispatch.message_received(message)

        if message.action == PROCESS_ADDED:
            self._process_added(message)
        elif message.action == PROCESS_REMOVED:
            self._process_closed(message)
        elif message.action == SUMMARY:
            self._summary_added(message)
        else:
            hostprog_key = message.hostname, message.progname
            conn_key = "%s|%s|%s" % (
                message.hostname, message.pid, message.connection_id)

            hostprog = self.hostprogs[hostprog_key]

            if message.action == CONNECT:
                hostprog.connect(
                    ConnectionStat(
                        message.hostname, message.pid, message.connection_id,
                        False, message.progname)
                )
            else:
                try:
                    stat = hostprog.connections[conn_key]
                except KeyError:
                    log.debug(
                        "Received message for unhandled connection %s; %s",
                        conn_key, message
                    )
                    assert conn_key not in hostprog.checkedout
                else:
                    if message.action == INVALIDATE:
                        hostprog.invalidate(stat)
                    elif message.action == CHECKOUT:
                        hostprog.checkout(stat)
                    elif message.action == CHECKIN:
                        hostprog.checkin(stat)
                    elif message.action == CLOSE:
                        hostprog.close(stat)


class QueuedStats(Stats):
    def __init__(self, _green=False):
        super(QueuedStats, self).__init__(_green=_green)
        self.listeners = set()
        self.queue = self.async_suite.queue()
        self.super_ = super(QueuedStats, self)
        self.receive_mutex = self.async_suite.lock()

    def start(self):
        self.runner = self.async_suite.background_thread(self._process_queue)

    def _process_queue(self):
        while True:
            message = self.queue.get()
            self.super_.receive_message(message)

    def receive_message(self, message):
        self.queue.put(message)


class StatsListener(util.EventListener):
    _dispatch_target = Stats

    def message_received(self, message):
        pass




from setuptools import setup
from setuptools import find_packages
import os
import re


with open(
    os.path.join(os.path.dirname(__file__), 'connmon', '__init__.py')
) as file_:
    VERSION = re.compile(
        r".*__version__ = '(.*?)'", re.S).match(file_.read()).group(1)


readme = os.path.join(os.path.dirname(__file__), 'README.rst')

requires = [
    'SQLAlchemy>=1.0.5',
    'eventlet',
    'zkcluster>=0.0.3'
]

setup(
    name='connmon',
    version=VERSION,
    description="Analyze database connection usage",
    long_description=open(readme).read(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Database :: Front-Ends',
    ],
    keywords='SQLAlchemy',
    author='Mike Bayer',
    author_email='mike@zzzcomputing.com',
    url='http://bitbucket.org/zzzeek/connmon',
    license='MIT',
    packages=find_packages('.', exclude=['examples*', 'test*']),
    include_package_data=True,
    tests_require=['mock'],
    zip_safe=False,
    install_requires=requires,
    entry_points={
        'console_scripts': [
            'connmon = connmon.frontend:monitor',
            'connmond = connmon.frontend:connmond',
        ],
        'sqlalchemy.dialects': [
            'mysql.connmon = connmon.plugin:CMWrapper',
            'mysql.mysqldb_connmon = connmon.plugin:CMWrapper',
            'mysql.pymysql_connmon = connmon.plugin:CMWrapper',
            'sqlite.connmon = connmon.plugin:CMWrapper',
            'postgresql.connmon = connmon.plugin:CMWrapper',
            'postgresql.psycopg2_connmon = connmon.plugin:CMWrapper',
        ]
    }
)
